#
#   Copyright 2012-2013 Erik Leunissen
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


# reval.tcl
#
# Implementation of the remote evaluation client
#

namespace eval ::reval {}


# ::reval::init --
#
# Initializes the package
#
# Arguments: none
#
# Result: none
#
proc ::reval::init {} {
	variable Reval
	variable Dir
	
	package require cmdline
	
	set Reval(patchLevel) 1.0.0
	
	# determine the installation directory
	set scriptname [info script]
	if {$scriptname eq ""} {
		return -code error "Called [info level 0] outside the active invocation of a script."
	}
	while {[file type $scriptname] eq "link"} {
		set scriptname [file readlink $scriptname]
	}
	set Dir(install) [file dirname [file normalize $scriptname]]
	
	# default settings; may be changed by reval::configure.
	set Reval(verbose) 0
	set Reval(host) localhost
	set Reval(port) 53423
	set Reval(timeout) 10000
	set Reval(mode) safe
	
	# trim patch level from the version number
	regexp {^([0-9]+\.[0-9]+)[.ab].*} $Reval(patchLevel) dummy Reval(version)
	package provide reval $Reval(version)
	namespace export reval
}


# ::reval::cleanup --
#
# Closes the channel and cleans up associated state
#
# Arguments
#     sock: the channel to the client
#
# Result: none
#
# Side effects: see description
#
proc ::reval::cleanup {sock} {
	variable Reval
	variable Channels
	variable Conn$sock
	upvar 0 Conn$sock conn
	
	if {[info exists conn(after)]} {
		after cancel $conn(after)
	}
	catch {close $sock}
	set session [lookupSession $sock]
	if {$session ne ""} {
		unset Channels($session)
	}
	if {[info exists Channels(current)] && ($sock eq $Channels(current))} {
		unset Channels(current)
	}
	unset conn
	
	if {$Reval(verbose)} {
		puts "closed channel $sock and cleaned up associated state"
	}
}


# ::reval::configure --
#
# Configure the default settings for the package
#
# Arguments: one or more key-value pairs
#
# Results: none
#
# Side effects: see description
#
proc ::reval::configure {args} {
	variable Reval
	variable Channels
	
	set argc [llength $args]
	set options {host mode port timeout verbose}
	if {$argc == 0} {
		foreach option $options {
			lappend result -$option $Reval($option)
		}
		return $result
	} elseif {$argc == 1} {
		set option [lindex $args 0]
		if {[string index $option 0] ne "-"} {
			return -code error "invalid option $option"
		} else {
			set option [string range $option 1 end]
		}
		if {[lsearch $options $option] < 0} {
			return -code error "invalid option $option"
		}
		
		return $Reval($option)
	} elseif {[string range [expr $argc / 2.] end-1 end] ne ".0"} {
		return -code error "[lindex [info level 0] 0] expects zero, one or an even nr. of arguments"
	}
	
	# The caller is trying to modify one or more settings. Validate the
	# option values.
	set options [cmdline::getoptions args { \
		{host.arg "" {machine to connect to}} \
				{mode.arg "" {evaluation mode}} \
				{port.arg "" {TCP port nr. to connect to}} \
				{timeout.arg "" {timeout value}} \
				{verbose.arg "" {be verbose about communication with service}} \
			} "[lindex [info level 0] 0] options:"]
	
	set sessionChange 0
	foreach {key value} $options {
		if {$value eq ""} {
			continue
		}
		switch -- $key {
			host {
				set sessionChange 1
			}
			port	{
				# -port argument ought be an integer
				if {![regexp {^[1-9][0-9]*$} $value]} {
					return -code error "invalid value \"$value\" for option -port"
				}
				set sessionChange 1
			}
			timeout {
				# -timeout argument ought be zero or a positive integer
				if {![regexp {^\+?[0-9][0-9]*$} $value]} {
					return -code error "invalid value \"$value\" for option -timeout"
				}
			}
			verbose {
				if {![string is boolean $value]} {
					return -code error "invalid value \"$value\" for option -verbose"
				}
			}
		}
		set Reval($key) $value
	}
	
	if {$sessionChange && [info exists Channels(current)]} {
		set session $Reval(host):$Reval(port)
		if {[info exists Channels($session)] && \
					($Channels($session) eq $Channels(current))} {
			return
		}
		
		# The newly configured host:port combination doesn't correspond to any
		# existing session. Unset Channels(current) to indicate that there is
		# no current session.
		unset Channels(current)
	}
}


# ::reval::connect --
#
# Opens a socket to the eval service revald.
#
# Arguments: none
#
# Results: 0 for failure, 1 for succes
#
# Side effects: see description
#
proc ::reval::connect {host port} {
	variable Reval
	variable Channels
	
	set session $host:$port
	if {[info exists Channels($session)]} {
		return $Channels($session)
	}

	# Open a socket to the remote eval service and wait for connection.
	# We do this asynchronously to allow for a connection timeout. This
	# prevents that the connection setup blocks for too long a time, e.g.
	# when a firewall blocks the connection attempt.
	set sock [socket -async $host $port]

	# store connection properties in a namespace array Conn$sock
	variable Conn$sock
	upvar 0 Conn$sock conn
	set conn(host) $host
	set conn(port) $port
	set conn(async) ""
	set conn(status) unknown
	set conn(state) init
	
	# prepare for connect timeout
	set conn(connect_timeout) 0
	set conn(after) [after $Reval(timeout) \
		[list set ::reval::Conn$sock\(connect_timeout) 1]]
	fileevent $sock writable [list ::reval::handleConnect $sock]
	vwait ::reval::Conn$sock\(connect_timeout)
	
	if {[info exists conn(connect_timeout)]} {
		cleanup $sock
		return -code error "connection timed out ($session)"
	} elseif {$conn(status) eq "error"} {
		set result $conn(result)
		cleanup $sock
		return -code error "$result ($session)"
	}
	
	# connection is ok, register session
	set conn(status) ok
	set conn(state) sync
	set Channels($session) $sock
	
	# wait for synchronization info from the server
	if {$Reval(timeout) > 0} {
		set conn(after) [after $Reval(timeout) [list ::reval::handleTimeout $sock]]
	}
	fileevent $sock readable [list ::reval::handleRead $sock]
	vwait ::reval::Conn$sock\(wait)
	
	if {$conn(status) eq "error"} {
		set code $conn(code)
		set result "while synchronizing with service: $conn(result)"
		cleanup $sock
		return -code $code $result
	}
	
	if {$Reval(verbose)} {
		puts "connection $sock to $session established"
	}
	set conn(free) 1
	return $sock
}


# ::reval::handleConnect --
#
# handler for connection accept
#
# Arguments:
#	sock: the socket to the service
#
# Results: none
#
# Side effects: see description
#
proc ::reval::handleConnect {sock} {
	variable Conn$sock
	upvar 0 Conn$sock conn
	
	# cancel the scheduled connect timeout
	after cancel $conn(after); unset conn(after)
	fileevent $sock writable {}
	unset conn(connect_timeout)
	set errMsg [fconfigure $sock -error]
	if {$errMsg ne ""} {
		handleDisconnect $sock fail_connect $errMsg
		return
	}
	
	fconfigure $sock -blocking 0
}


# ::reval::handleDisconnect --
#
# handler for an error status
#
# Arguments:
#	sock: the socket to the service
#	reason: the reason for disconnecting
#	errMsg: an error message
#
# Results: none
#
# Side effects: closes the socket or flags an error status to let [reval]
#               do that
#
proc ::reval::handleDisconnect {sock reason {errMsg ""}} {
	variable Conn$sock
	upvar 0 Conn$sock conn
	
	if {$conn(state) ne "init"} {
		set msg "reval session [lookupSession $sock]: "
	}
	switch -- $reason {
		eof {
			append msg "the connection to the service broke unexpectedly (eof), while "
		}
		fail_connect -
		readError {
			append msg "$errMsg, while "
		}
		serviceShutdown {
			append msg "the remote service was shut down"
		}
	}
	switch -- $conn(state) {
		init {
			append msg "setting up the connection to the service"
		}
		sync {
			append msg "synchronizing with the service"
		}
		waitingForResult {
			append msg "waiting for the result to arrive"
		}
		result {
			append msg "reading the result"
		}
		done {
			if {$reason ne "serviceShutdown"} {
				error "unexpected reason $reason in state $conn(state)"
			}
		}
		default {
			error "unexpected state $conn(state)"
		}
	}
	
	fileevent $sock readable {}
	fileevent $sock writable {}
	if {$conn(state) eq "done"} {
		if {$::tcl_interactive} {
			puts stderr "warning: $msg"
		}
		cleanup $sock
	} else {
		set conn(status) error
		set conn(code) 1
		set conn(result) $msg
		set conn(errorInfo) ""
		set conn(errorCode) ""
		routeResult $sock
	}
}


# ::reval::handleRead --
#
# Read handler for the response from revald.
#
# Arguments:
#	sock: the socket to the service
#
# Results: none
#
# Side effects: see description
#
proc ::reval::handleRead {sock} {
	variable Reval
	variable Channels
	variable Conn$sock
	upvar 0 Conn$sock conn
	
	if {[info exists conn(after)]} {
		# cancel the scheduled timeout for the response from the service
		after cancel $conn(after); unset conn(after)
	}
	
	if {[catch {gets $sock line} errMsg]} {
		if {$Reval(verbose)} {
			puts "read error on $sock: $errMsg"
		}
		handleDisconnect $sock readError $errMsg
		return
	}
	
	if {[eof $sock]} {
		if {$Reval(verbose)} {
			puts "eof on $sock"
		}
		handleDisconnect $sock eof
		return
	}
	
	if {$Reval(verbose)} {
		puts "read $sock: $line"
	}
	
	if {$line eq "REVALD_SHUTDOWN"} {
		handleDisconnect $sock serviceShutdown
		return
	}
	
	# process the line
	switch -- $conn(state) {
		sync {
			# read synchronization info
			if {$line eq "REVALD_DONE"} {
				set conn(state) done
				fileevent $sock readable {}
				set conn(wait) 0
				return
			}
			if {[string match REVALD_MODES* $line]} {
				if {![info exists conn(modes)]} {
					set conn(modes) [string tolower [string range $line 13 end]]
				}
				return
			}
			if {[regexp {^REVALD_IDLE_TIMEOUT (\d+)$} $line dummy value]} {
				set conn(idle_timeout) [expr {1000*$value}]
				return
			}
			error "while in state $conn(state): unexpected line: $line"
		}
		waitingForResult {
			if {$line eq ""} {
				# This may happen on Linux if the revald service was forcibly
				# killed. An eof will follow, and handle the case.
				return
			}
			if {[regexp {^REVALD_CODE ([+-]?\d+)$} $line dummy conn(code)]} {
				# the evaluation result arrived
				set conn(state) result
			} else {
				error "while in state $conn(state): unexpected line: $line"
			}
		}
		result {
			switch -glob -- $line {
				REVALD_DONE* {
					set conn(state) done
					fileevent $sock readable {}
					if {$line eq "REVALD_DONE_AND_CLOSE"} {
						#
						# The service has closed the socket after having 
						# evaluated a script that calls [exit]. We indicate
						# here to [reval] to call [cleanup].
						# 
						set conn(status) close
					} elseif {$conn(code) == 1} {
						error "wrong return code $conn(code)"
					}
					
					#
					# The service has already started counting idle connection
					# time. We will let [reval] handle that case.
					#
					
					set conn(result) [string range $conn(result) 0 end-1]
					routeResult $sock
					return
				}
				REVALD_ERRORINFO {
					set conn(result) [string range $conn(result) 0 end-1]
					set conn(state) errorInfo
					return
				}
			}
			append conn(result) $line
			append conn(result) \n
		}
		errorInfo {
			if {[string match "REVALD_ERRORCODE" $line]} {
				set conn(errorInfo) [string range $conn(errorInfo) 0 end-1]
				set conn(state) errorCode
				return
			}
			append conn(errorInfo) $line
			append conn(errorInfo) \n
		}
		errorCode {
			if {[string match "REVALD_DONE*" $line]} {
				if {$line eq "REVALD_DONE_AND_CLOSE"} {
					#
					# The server closes the socket when done.
					# This happens when:
					# - the client supplied an invalid evaluation mode
					# - other irrecoverable server errors
					#
					set conn(status) close
				}
				set conn(state) done
				fileevent $sock readable {}
				set conn(errorCode) [string range $conn(errorCode) 0 end-1]
				routeResult $sock
				return
			}
			append conn(errorCode) $line
			append conn(errorCode) \n
		}
		done {
			if {$line eq ""} {
				# This may happen in between successive reval commands
				# on Linux if the revald service was forcibly killed.
				# An eof will follow, and handle the case.
				return
			}
			error "while in state $conn(state): unexpected line: $line"
		}
		default {
			error "while in state $conn(state): unexpected line: $line"
		}
	}
}


# ::reval::handleTimeout --
#
# Closes the socket
#
# Arguments
# - sock: channel to the service
#
# Result: none
#
# Side effects: see description
#
proc ::reval::handleTimeout {sock} {
	variable Conn$sock
	upvar 0 Conn$sock conn
	variable Reval
	
	set conn(status) error
	set conn(result) "remote evaluation timed out (session [lookupSession $sock])"
	set conn(code) 1
	set conn(errorInfo) $conn(result)
	set conn(errorCode) $::errorCode
	routeResult $sock
}


# ::reval::handleIdleTimeout --
#
# Closes the socket
#
# Arguments:
#	sock: the socket to handle
#
# Result: none
#
# Side effects:
#	closes the socket
#	writes a notification to stderr
#
proc ::reval::handleIdleTimeout {sock} {
	variable Reval
	variable Channels
	variable Conn$sock
	upvar 0 Conn$sock conn
	
	# catch this because the service may already have closed the socket
	catch {
		puts $sock "REVAL_IDLE_TIMEOUT_CLOSE"
		flush $sock
	}
	if {[lsearch [file channels] stderr] >= 0} {
		set limit [expr {$conn(idle_timeout) / 60000.0}]
		set unit minute
		if {$limit > 1.0} {
			append unit s
		}
		puts stderr "the session with revald service at \"[lookupSession $sock]\": exceeded the idle time limit ($limit $unit);\
				\nthe next evaluation will use a new execution environment."
	}
	cleanup $sock
}


# ::reval::isfree --
#
# Indicates whether the current connection is free. To be used when
# scripts are processed asynchronously.
#
# Arguments: none
#
# Result: 0 or 1
#
# Side effects: none
#
proc ::reval::isfree {} {
	variable Channels
	
	if {! [info exists Channels(current)]} {
		return 1
	}
	
	variable Conn$Channels(current)
	upvar 0 Conn$Channels(current) conn
	return $conn(free)
}


proc ::reval::lookupSession {sock} {
	variable Channels
	
	# find session amongst registered sessions
	foreach session [array names Channels] {
		if {$sock eq $Channels($session)} {
			return $session
		}
	}
}


proc ::reval::reval {args} {
	variable Reval
	
	set argc [llength $args]
	if {$argc < 1} {
		return -code error "wrong #args. Usage: [lindex [info level 0] 0] ?options? arg"
	}
	
	set opts [lrange $args 0 end-1]
	set options [cmdline::getoptions opts { \
		{async.arg "" {return the result asynchronously} } \
				{timeout.arg "" {timeout value}} \
			} "[lindex [info level 0] 0] options:"]
	
	foreach {key value} $options {
		set $key $value
	}
	
	# validate the option values
	if {$timeout ne ""} {
		# -timeout argument ought be an integer >= 0
		if {![regexp {^\+?\d+$} $timeout]} {
			return -code error "invalid value $timeout for option -timeout"
		}
	}
	if {($async ne "") && \
				[llength [uplevel #0 [list info commands [lindex $async 0]*]]] == 0} {
		return -code error "command \"$async\" doesn't exist"
	}
	
	# verify that the connection exists; reconnect if it doesn't
	variable Channels
	set reconnect 1
	if {[info exists Channels(current)]} {
		set sock $Channels(current)
		variable Conn$sock
		upvar 0 Conn$sock conn
		
		# check whether the channel is free
		if {!$conn(free)} {
			return -code error "channel to [lookupSession $sock] is busy"
		}
		set reconnect 0
		
		#
		# Assume that the current channel is in order. There is no 100%
		# guarantee for that though, but the probability of exceptions[*]
		# is considered too low to merit a check here.
		#
		# [*] reval.test provokes such an exceptional case by
		# utilizing [after <seconds>]
		#
	}
	
	if {$reconnect} {
		if {[catch {
				connect $Reval(host) $Reval(port)
			} sock]} {
			return -code error $sock
		}
		set Channels(current) $sock
		variable Conn$sock
		upvar 0 Conn$sock conn
	}
	
	# validate the evaluation mode for this connection
	if {[lsearch $conn(modes) $Reval(mode)] < 0} {
		return -code error "invalid evaluation mode \"$Reval(mode)\""
	}
	set conn(mode) $Reval(mode)
	
	# assign defaults to the connection state array unless
	# overridden by the command line (except for -async)
	foreach {key value} $options {
		if {($key eq "async") || ($value ne "")} {
			set conn($key) $value
		} else {
			set conn($key) $Reval($key)
		}
	}
	set conn(script) [lindex $args end]
	
	# send the script to the remote host
	if {[catch {
			send $sock
		} errMsg]} {
		return -code error $errMsg
	}
	
	if {$conn(async) ne ""} {
		# The non-blocking case: hand control over to the event loop, and
		# leave it to [handleRead] or [handleTimeout] to invoke the callback.
		return
	}
	
	# wait for the result to become available (blocking)
	vwait ::reval::Conn$sock\(wait)
	
	# capture the result
	set code $conn(code)
	set result $conn(result)
	if {$conn(code) == 1} {
		# adjust error info to reflect processing by the remote service
		regsub "while executing" $conn(errorInfo) "while remotely executing" errorInfo
		set errorCode $conn(errorCode)
	}
	
	# cleanup or schedule an idle timeout event
	if {($conn(status) eq "error") || ($conn(status) eq "close")} {
		cleanup $sock
	} else {
		# While the connection is kept alive, the service keeps listening for
		# a limited time, which it has communicated as the idle timeout limit.
		# Schedule an event as indicated by the service.
		set conn(after) [after $conn(idle_timeout) [list ::reval::handleIdleTimeout $sock]]
	}
	
	# return
	if {$code == 1} {
		return -code $code -errorinfo $errorInfo -errorcode $errorCode $result
	}
	return -code $code $result
}


# ::reval::routeResult --
#
# Routes the result to one of the following procs:
# - the result callback registered with the option -async, or
# - [reval], which is still waiting for the result to become available
#   in blocking mode.
#
# Arguments
# - sock: channel to the service
#
# Result: none
#
# Side effects: see description
#
proc ::reval::routeResult {sock} {
	variable Conn$sock
	upvar 0 Conn$sock conn
	
	set conn(free) 1
	if {$conn(async) eq ""} {
		# unblock vwait to let [reval] continue processing
		set conn(wait) 0
	} else {
		# pass the result on to the registered result callback
		set cmd $conn(async)
		lappend cmd $conn(code) $conn(result)
		if {$conn(code) == 1} {
			lappend cmd $conn(errorInfo)
			lappend cmd $conn(errorCode)
		}
		
		if {($conn(status) eq "error") || ($conn(status) eq "close")} {
			cleanup $sock
		}
		
		eval $cmd
	}
}


# ::reval::send --
#
# Sends a script to revald and schedules an event handler for reading the
# result from revald (plus timeout handler)
#
proc ::reval::send {sock} {
	variable Reval
	variable Conn$sock
	upvar 0 Conn$sock conn
	
	# initialize the transaction
	unset -nocomplain conn(code) conn(result) conn(errorInfo) conn(errorCode)
	set conn(free) 0
	set conn(wait) 1
	
	puts $sock "REVAL_MODE $conn(mode)"
	puts $sock $conn(script)
	puts $sock REVAL_DONE
	flush $sock
	
	if {$Reval(verbose)} {
		puts "sent script to [lookupSession $sock]"
	}
	
	if {[info exists conn(after)]} {
		# cancel the scheduled idle timeout from the previous request
		after cancel $conn(after); unset conn(after)
	}
	fileevent $sock readable [list ::reval::handleRead $sock]
	if {$conn(timeout) > 0} {
		set conn(after) [after $conn(timeout) [list ::reval::handleTimeout $sock]]
	}
	set conn(state) waitingForResult
}


# ::reval::server_cget --
#
# Return the settings supported by the current remote service
#
# Arguments:
#    setting: one of "modes" or "idle_timeout"
#
# Results: a list of evaluation modes or an idle timeout value
#
# Side effects: connects if there wasn't any connection
#
proc ::reval::server_cget {setting} {
	variable Reval
	variable Channels
	
	set allowed [list modes idle_timeout]
	if {[lsearch $allowed $setting] < 0} {
		return -code error "invalid argument \"$setting\""
	}
	
	if {[info exists Channels(current)]} {
		variable Conn$Channels(current)
		upvar 0 Conn$Channels(current) conn
	} else {
		if {[catch {
				connect $Reval(host) $Reval(port)
			} sock]} {
			return -code error $sock
		}
		set Channels(current) $sock
		variable Conn$sock
		upvar 0 Conn$sock conn
	}
	return $conn($setting)
}


proc reval::terminate {} {
	variable Channels
	foreach session [array names Channels] {
		cleanup $Channels($session)
	}
	namespace delete ::reval
	package forget reval
}


# ::reval::version --
#
# Returns the version + patch level of this package
#
# Arguments: none
#
# Results: the version number
#
# Side effects: none
#
proc ::reval::version {} {
	return $::reval::Reval(patchLevel)
}


::reval::init

# EOF
