#! /bin/sh
# If executed as a shell script, the next line replaces the shell with tclsh \
exec tclsh "$0" ${1+"$@"}

#
#   Copyright 2012-2013 Erik Leunissen
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


# revald.tcl
#
# Implementation of the remote evaluation service
#

# Revald_Init --
#
# Initialize the service
#
# Arguments: none
#
# Results: none
#
# Side effects:
# - reads the configuration file
# - starts logging
# - sets initial values for variables
#
proc Revald_Init {} {
	variable	Dir
	variable	Revald
	
	set Revald(patchLevel) 1.0.0
	
	# trim patch level from the version number
	regexp {^[0-9]+\.[0-9]+} $Revald(patchLevel) Revald(version)
	
	#
	# Determine the absolute position of this script in the system's
	# directory tree.
	#
	proc GetAbsoluteScriptName {} {
		set scriptname [info script]
		
		if {$scriptname eq ""} {
			return -code error "Called [info level 0] outside the active invocation of a script."
		}
		while {[file type $scriptname] eq "link"} {
			set scriptname [file readlink $scriptname]
		}
		
		return [file normalize $scriptname]
	}
	set Dir(install) [file dirname [GetAbsoluteScriptName]]
	if {[file tail $Dir(install)] eq "src"} {
		# we're in a source tree; climb up one level
		set Dir(install) [file dirname $Dir(install)]
	}
	set Dir(evalmodes) [file join $Dir(install) evalmodes]
	rename GetAbsoluteScriptName {}
	
	# initialize configurable variables
	array set Revald {
		ALLOWED_HOSTS "127.0.0.1 ::1"
		COMPRESS_CMD gzip
		EVAL_MODES safe
		IDLE_TIMEOUT_SECONDS 600
		IP_ADDRESS localhost
		LOG_DIR /var/log/revald
		LOG_KEEP_DAYS 10
		PID_FILE /var/run/revald.pid
		PORT 53423
	}
	
	# read configuration file
	set ndx [lsearch $::argv "-config"]
	if {$ndx >= 0} {
		if {$::argc <= [incr ndx]} {
			return -code error "the option \"-config\" requires an additional argument"
		}
		set Revald(cfgFile) [file normalize [lindex $::argv $ndx]]
	} else {
		set Revald(cfgFile) [file join $Dir(install) revald.conf]
	}
	Revald_ReadCfgFile $Revald(cfgFile) Revald
	
	# compose configuration text for log file
	set config_info "Service configuration:\n"
	set keys [lsort [array names Revald]]
	foreach name $keys {
		append config_info "\t$name: $Revald($name)\n"
	}
	
	# start logging
	set Dir(log) $Revald(LOG_DIR)
	unset Revald(LOG_DIR)
	Revald_LogInit
	log::log notice "*** Program launch revald $Revald(patchLevel) (pid=[pid], executable = [info nameofexecutable])"
	log::log info $config_info
	
	# check for presence of scripts defining custom evaluation modes
	foreach mode $Revald(EVAL_MODES) {
		if {[lsearch {safe standard full} $mode] >= 0} {
			# these modes are predefined
			continue
		}
		if {![file isfile [file join $Dir(evalmodes) $mode.tcl]]} {
			set errMsg "couldn't find custom mode definition script\
					[file join $Dir(evalmodes) $mode.tcl]"
			log::log error $errMsg
			log::log notice "program halted (pid=[pid])"
			exit 1
		}
	}
	
	# initialize namespace variables
	set Revald(init) 1
	set Revald(shutdown) 0
	set Revald(timeout_write) 1000
	
	# Use an offset of 2 seconds to let the client process time out
	# just before the service closes an idle connection.
	set Revald(idle_timeout) [expr {1000*$Revald(IDLE_TIMEOUT_SECONDS) + 2000}]
}


# Revald_Accept --
#
# handles connections from remote clients
#
# Arguments
# - sock: the channel to the client
# - host: IP address of the peer
# - port: TCP port nr. at the peer's side
#
# Results: none
#
# Side effects:
# - configures the channel
# - returns synchronization info
# - creates a fresh execution environment
# - registers a handler for reading what the client sends
# - schedules a timeout handler in case of trouble
#
proc Revald_Accept {sock host port} {
	variable Revald
	
	log::log info "connection attempt from host $host, port $port"
	
	#
	# Host-based authentication is unreliable because IP spoofing can
	# work around it. Nevertheless, it may be useful for testing
	# purposes. Don't rely on it for regular operation.
	#
	set allowed 0
	foreach pattern $Revald(ALLOWED_HOSTS) {
		if {[string match $pattern $host]} {
			set allowed 1
			break
		}
	}
	if {! $allowed} {
		log::log notice "access denied to $host"
		catch {close $sock}
		return
	}
	
	# record the connection in a namespace variable
	upvar #0 Conn$sock conn
	set conn(state) init
	set conn(host) $host
	set conn(port) $port
	set conn(status) ok
	
	# configure the channel
	fconfigure $sock -encoding utf-8 -translation {auto crlf} \
			-buffering line -blocking 0
	
	if {![Revald_ReturnSyncInfo $sock]} {
		log::log notice "failed to sync connection $sock"
		Revald_Cleanup $sock
		return -code error
	}
	set conn(state) reset
	fileevent $sock readable [list Revald_Read $sock]
	if {$Revald(IDLE_TIMEOUT_SECONDS) > 0} {
		set conn(after) [after $Revald(idle_timeout) [list Revald_HandleTimeout $sock idle]]
	}
}


# Revald_Cleanup --
#
# Delete/remove everything related to a client connection
#
# Arguments
# - sock: the channel to the client
#
# Results: none
#
# Side effects:
# - deletes interpreter for the client connection
# - closes the channel and deletes any associated file or timer events
# - unsets the namespace variable for the connection
# - writes a notification into the log
#
proc Revald_Cleanup {sock} {
	variable Revald
	upvar #0 Conn$sock conn
	
	if {[interp exists $sock]} {
		interp delete $sock
	}
	catch {fileevent $sock readable {}}
	catch {fileevent $sock writable {}}
	catch {close $sock}
	if {[info exists conn(after)]} {
		# cancel any timer events
		after cancel $conn(after)
	}
	if {$conn(state) eq "reset"} {
		set msg "closed channel $sock and cleaned up associated state"
	} else {
		set msg "forcefully closed channel $sock while in state \"$conn(state)\", and cleaned up associated state"
	}
	log::log notice $msg
	unset conn
}


# Revald_ClientExit --
#
# This proc is a replacement for [exit].
#
# Arguments
# - sock: the channel to the client
# - exitStatus: the code supplied to the original exit command.
#               The exit status is returned to the client as
#               the evaluation result (default is an empty string).
#               This makes [exit] behave as [return]
#
# Results: none
#
# Side effects:
# - writes a notification into the log
# - records the call to exit in conn(status), to handle it after having
#   returned the result to the client.
#
proc Revald_ClientExit {sock {exitStatus ""}} {
	upvar #0 Conn$sock conn
	
	log::log notice "script for channel $sock called exit"
	set conn(status) exit
	
	#
	# We raise an (artificial) error in order to abort script execution.
	# Since this "error" isn't caused by script evaluation, we need
	# to restore the ok return code later.
	#
	# TODO #10:
	# Replace this hack by [interp cancel] as soon as that is available
	# (in Tcl8.6, see also TIP #285)
	#
	return -code error $exitStatus
}


# Revald_CreateInterp --
#
# Creates an interpreter as the execution environment for a script
#
# Arguments
# sock: the channel to the client
#
# Results: none
#
# Side effects: see description
#
proc Revald_CreateInterp {sock} {
	variable Dir
	upvar #0 Conn$sock conn
	
	#
	# We simply name dedicated interpreters after the channel to the client
	#
	if {[interp exists $sock]} {
		return
	}
	
	#
	# TODO
	# Authorization for evaluation modes other than "safe". This requires
	# that the connection is encrypted (TLS).
	#
	# - check whether the requested evaluation mode has been
	#   configured/installed with Revald_EvalModeInstall.
	# - check whether the remote host/user is authorized
	#   for that specific mode.
	#
	#   Revald_Auth $sock
	#
	
	switch -- $conn(mode) {
		full {
			#
			# like "standard", but enhanced with a copy of argv0, argc
			# and argv, to mimic a main interpreter.
			#
			interp create $sock
			interp eval $sock [list set argv0 $::argv0]
			interp eval $sock [list set argc $::argc]
			interp eval $sock [list set argv $::argv]
		}
		safe {
			interp create -safe $sock
		}
		standard {
			#
			# A regular interpreter, but different from a main interpreter.
			# It lacks argv0, argc, and argv, and has tcl_interactive set to 0.
			#
			interp create $sock
		}
		default {
			#
			# Source the definition script for the custom evaluation mode.
			#

			#
			# TODO: make this more secure (it's now horribly unsafe!)
			#
			# We don't need to worry about invalid evaluation modes,
			# requested by the client since these have already been
			# captured by Revald_Read.
			#
			source [file join $Dir(evalmodes) $conn(mode).tcl]
			if {![info exists interpCreateCmd]} {
				return -code error "mode definition script \"[file join $Dir(evalmodes) $conn(mode).tcl]\" \
					doesn't define a command \"interpCreateCmd\" to create a dedicated interpreter"
			}
			eval [lappend interpCreateCmd $sock]
			if {![interp exists $sock]} {
				return -code error "mode definition script \"[file join $Dir(evalmodes) $conn(mode).tcl]\" \
					failed to create a dedicated interpreter"
			}
			
			if {[info exists customizationCode]} {
				interp eval $sock $customizationCode
			}
		}
	}
	
	# prevent that the client script terminates the program
	interp alias $sock exit {} Revald_ClientExit $sock
	log::log info "created $conn(mode) interp for channel $sock"
}


# Revald_Evaluate --
#
# evaluates a script, sent by a remote client
#
# Arguments:
# - sock: the channel to the client
#
# Results: none
#
# Side effects:
# - anything the script does
# - collects return code, result and error info from the script evaluation
# - logs any interp creation, and completion of the script evaluation
#
proc Revald_Evaluate {sock} {
	variable Revald
	upvar #0 Conn$sock conn
	
	if {[catch {
			Revald_CreateInterp $sock
		} errMsg]} {
		set conn(code) 1
		set conn(result) "Revald service error: failed to create a dedicated interpreter. \
		Internal error message: $errMsg"
		set conn(errorInfo) $::errorInfo
		set conn(errorCode) "REVALD_FAIL_INTERP_CREATE"
		set conn(status) error
		log::log error "(for channel $sock) $errMsg"
		log::log info "did not evaluate script for channel $sock"
		return
	}
	
	set conn(code) [catch {
		interp eval $sock $conn(script)
	} conn(result)]
	
	if {$conn(status) eq "exit"} {
		#
		# A script called [exit], Revald_ClientExit raised an
		# error. This error is artificial because its only purpose is
		# to abort script evaluation. Restore the ok return code here.
		#
		set conn(code) 0
	}
	
	if {$conn(code) == 1} {
		set conn(errorInfo) [interp eval $sock set ::errorInfo]
		# syntax errors for the exit command ought not reveal the alias
		if {[string match "*Revald_ClientExit*" $conn(errorInfo)]} {
			regsub "Revald_ClientExit sock" $conn(errorInfo) exit conn(errorInfo)
			regsub "Revald_ClientExit sock" $conn(result) exit conn(result)
		}
		set conn(errorCode) [interp eval $sock set ::errorCode]
	}
	log::log info "done evaluating script for channel $sock"
}


# Revald_HandleTimeout --
#
# handles various timeout events
#
# Arguments
# - sock: the channel to the client
# - type: the type of timeout event
#
# Results: none
#
# Side effects:
# - writes a notification into the log
# - cleans up the connection to the client
#
proc Revald_HandleTimeout {sock type} {
	switch -- $type {
		idle {
			fileevent $sock readable {}
		}
		writeresult {
			fileevent $sock writable {}
		}
		default {
			error "unknown timeout type $type (for channel $sock)"
		}
	}
	log::log notice "$type timeout for channel $sock"
	Revald_Cleanup $sock
}


# Revald_Read --
#
# handles readable events on the socket
#
# Arguments
# sock: the channel to the client
#
# Results: none
#
# Side effects
# - various side effects (depending on the read), notably script evaluation
#
proc Revald_Read {sock} {
	variable Revald
	upvar #0 Conn$sock conn
	
	# read line
	if {[catch {gets $sock line} errMsg] || [eof $sock]} {
		#
		# The channel to the client was broken
		#
		if {$errMsg == -1} {
			log::log notice "unexpected eof on socket $sock (maybe ungraceful client exit)"
		} else {
			log::log error $::errorInfo
		}
		Revald_Cleanup $sock
		return
	}
	
	switch -- $conn(state) {
		reset {
			if {[info exists conn(after)]} {
				# cancel any idle timeout event
				after cancel $conn(after)
				unset conn(after)
			}
			switch -glob -- $line {
				"REVAL_MODE *" {
					regexp {^REVAL_MODE (.*)$} $line dummy mode
					if {[lsearch $Revald(EVAL_MODES) $mode] < 0} {
						set conn(status) error
						set conn(code) 1
						set conn(result) "invalid evaluation mode \"$mode\""
						set conn(errorInfo) $conn(result)
						set conn(errorCode) "REVALD_INVALID_MODE"
						log::log error "(for channel $sock) $conn(result)"
						
						# Although we're not going to evaluate anything, we
						# need to consume what's in the channel to prevent
						# communication errors. The script will be discarded.
						set conn(state) readScript
						return
					}
					if {[info exists conn(mode)] && ($mode ne $conn(mode))} {
						#
						# client changed evaluation mode. The current
						# interp isn't valid anymore. Delete it.
						#
						log::log notice "client changed eval mode to \"$mode\" (channel $sock)"
						if {[interp exists $sock]} {
							interp delete $sock
							log::log info "deleted $conn(mode) interp for channel $sock"
						}
					}
					set conn(mode) $mode
					set conn(state) readScript
				}
				"REVAL_IDLE_TIMEOUT_CLOSE" {
					log::log info "received disconnect notice from $sock"
					Revald_Cleanup $sock
				}
				default {
					log::log error "bad initial line $line, read from $sock"
					Revald_Cleanup $sock
				}
			}
		}
		readScript {
			if {$line eq "REVAL_DONE"} {
				log::log info "received script on channel $sock"
				fileevent $sock readable {}
				set conn(script) [string range $conn(script) 0 end-1]
				if {$conn(status) eq "ok"} {
					set conn(state) evaluating
					Revald_Evaluate $sock
				} else {
					log::log info "did not evaluate script for channel $sock"
				}
				unset conn(script)
				fileevent $sock writable [list Revald_ReturnResult $sock]
				set conn(after) [after $Revald(timeout_write) [list Revald_HandleTimeout $sock writeresult]]
				vwait Conn$sock\(status)
				
				set conn(state) reset
				if {($conn(status) eq "error") || ($conn(status) eq "exit") \
							|| $Revald(shutdown)} {
					if {$Revald(shutdown)} {
						puts $sock REVALD_SHUTDOWN
						flush $sock
					}
					Revald_Cleanup $sock
					if {$Revald(shutdown)} {
						if {[llength [info globals Conn*]] == 0} {
							after cancel $Revald(after)
							set Revald(shutdown) 2
						}
					}
					return
				}
				
				# prepare for reading the next evaluation request
				fileevent $sock readable [list Revald_Read $sock]
				if {$Revald(IDLE_TIMEOUT_SECONDS) > 0} {
					set conn(after) [after $Revald(idle_timeout) [list Revald_HandleTimeout $sock idle]]
				}
			} else {
				append conn(script) $line
				append conn(script) \n
			}
		}
		default {
			log::log error "unexpected read from channel $sock while in state $conn(state):\
					$line"
		}
	}
}


# Revald_ReadCfgFile --
#
# Process the configuration file
#
# Arguments:
#   cfgFile : the file to read the settings from
#   varName : the name of an array variable to collect the settings in
#
# Results:
#    the empty string
#
# Side effects:
#    Sets entries in the array given by varName
#
proc Revald_ReadCfgFile {cfgFile varName} {
	upvar #0 $varName configVar
	
	set fd [open $cfgFile r]
	set i 0
	while { ([gets $fd line] >= 0) && ![eof $fd] } {
		incr i
		if {[string match #* $line] || \
					[string is space $line]} {
			# note that [string is space $line] also
			# matches the empty string.
			continue
		}
		set re {^\s*(.+?)\s*[:=]\s*(.*?)\s*$}
		if {[regexp $re $line dummy key value]} {
			if {![info exists configVar($key)]} {
				return -code error "invalid configuration key \"$key\" in line $i"
			}
			if {[string length $value] == 0} {
				set configVar($key) {}
				continue
			}
			set configVar($key) $value
		} else {
			close $fd
			error "Error parsing line $i:\n$line"
		}
	}
	close $fd
}


# Revald_Restart --
#
# Restarts the service, rereading the configuration file
#
# Arguments: none
#
# Results: none
#
# Side effects: see description
#
proc Revald_Restart {} {
	global Revald
	
	Revald_Stop
	Revald_ReadCfgFile $Revald(cfgFile) Revald
	Revald_Start
}


# Revald_ReturnSyncInfo --
#
# Returns synchronization info to the client:
# - a list of the supported evaluation modes
# - the idle timeout value, to allow the client to time out before
#   the server does so in an ungraceful manner.
#
# Arguments
# sock: the channel to the client
#
# Results: 0 for failure, 1 for success
#
# Side effects: see description
#
proc Revald_ReturnSyncInfo {sock} {
	variable Revald
	upvar #0 Conn$sock conn
	
	if {[catch {
			puts $sock "REVALD_MODES $Revald(EVAL_MODES)"
			puts $sock "REVALD_IDLE_TIMEOUT $Revald(IDLE_TIMEOUT_SECONDS)"
			puts $sock REVALD_DONE
			flush $sock
		} errMsg]} {
		log::log error $::errorInfo
		set conn(status) error
		return 0
	}
	log::log info "returned synchronization info for channel $sock"
	return 1
}


# Revald_ReturnResult --
#
# Returns the evaluation result to the client
#
# Arguments
# sock: the channel to the client
#
# Results: none
#
# Side effects: see description
#
proc Revald_ReturnResult {sock} {
	upvar #0 Conn$sock conn
	
	after cancel $conn(after)
	unset conn(after)
	if {[catch {
			puts $sock "REVALD_CODE $conn(code)"
			puts $sock $conn(result)
			if {($conn(code) == 1) && ($conn(status) ne "exit")} {
				puts $sock REVALD_ERRORINFO
				puts $sock $conn(errorInfo)
				puts $sock REVALD_ERRORCODE
				puts $sock $conn(errorCode)
			}
			set str REVALD_DONE
			if {($conn(status) eq "exit") || ($conn(status) eq "error")} {
				append str _AND_CLOSE
			}
			puts $sock $str
			flush $sock
		} errMsg]} {
		log::log error $::errorInfo
		set conn(status) error ;# set waitVar
		return
	}
	
	log::log info "returned result for channel $sock"
	fileevent $sock writable {}
	set conn(status) $conn(status) ;# set waitVar, but don't change its value!
}


# Revald_Shutdown --
#
# Stops the service and exits
#
# Arguments:
# seconds: the service is halted after seconds, regardless the
#          completion of any client requests
#
# Results: none
#
proc Revald_Shutdown {seconds} {
	variable Revald
	
	set Revald(shutdown) 1
	log::log info "initiated service shutdown"
	
	Revald_Stop
	
	# close all idle connections
	set openConnections [info globals Conn*]
	set count [llength $openConnections]
	foreach varName $openConnections {
		set sock [string range $varName 4 end]
		upvar #0 Conn$sock conn
		if {$conn(state) eq "reset"} {
			puts $sock REVALD_SHUTDOWN
			flush $sock
			Revald_Cleanup $sock
			incr count -1
		}
	}
	
	# wait for transactions in progress to complete
	if {$count > 0} {
		set Revald(after) [after [expr {1000*$seconds}] {set Revald(shutdown) 2}]
		log::log info "waiting $seconds seconds for all current transactions to complete ..."
		vwait Revald(shutdown)
		
		# forcefully close all remaining connections
		set openConnections [info globals Conn*]
		foreach varName $openConnections {
			set sock [string range $varName 4 end]
			puts $sock REVALD_SHUTDOWN
			flush $sock
			Revald_Cleanup $sock
		}
	}
	
	# halt
	log::log notice "Program halted (pid=[pid])"
	catch {close $Revald(log_fd)}
	catch {file delete $Revald(PID_FILE)}
	exit 0
}


# Revald_Start --
#
# Starts the service
#
# Arguments: none
#
# Results: none
#
# Side effects:
# - opens a listening socket
# - logs the event
#
proc Revald_Start {} {
	variable Revald
	
	if {![info exists Revald(init)] || ($Revald(init) != 1)} {
		Revald_Init
	}
	
	# protect from glaring security risks
	if {($::tcl_platform(user) eq "root") || [string match -nocase admin* $::tcl_platform(user)]} {
		return -code error "Attempt to run revald with administrator \
				permissions (user \"$::tcl_platform(user)\"). That is too big \
				a security risk. Will exit now."
		after 2000
		exit 1
	}
	
	set cmd [list socket -server Revald_Accept -myaddr $Revald(IP_ADDRESS) $Revald(PORT)]
	if {[catch $cmd Revald(listenSock)]} {
		return -code error $Revald(listenSock)
	}
	
	if {[info exists Revald(PID_FILE)] && ($Revald(PID_FILE) ne "")} {
		if {[catch {
			set fd [open $Revald(PID_FILE) w]
			puts $fd [pid]
			close $fd
		} errMsg]} {
			return -code error $errMsg
		}
	}
	
	set msg "started listening at $Revald(IP_ADDRESS), port $Revald(PORT)"
	log::log notice $msg
	append msg "\n\tconfig file : $Revald(cfgFile)\n\tlog file    : $Revald(logfile)"
	catch {puts $msg}
}


# Revald_Stop --
#
# Stops accepting new connections from clients, but
# keep existing connections and corresponding interpreters.
#
# Arguments: none
#
# Results: none
#
# Side effects:
# - closes the listening socket
# - writes a notification into the log
#
proc Revald_Stop {} {
	global Revald
	catch {close $Revald(listenSock)}
	unset Revald(listenSock)
	log::log notice "revald stopped listening for new connections"
}


# Revald_Version --
#
# Return the patch level
#
# Arguments: none
#
# Results:
#    the patch level
#
# Side effects: none
#
proc Revald_Version {} {
	return ::Revald(patchLevel)
}


# ::bgerror --
#
# handles background errors (writing them to the log file)
#
# Arguments:
# errMsg: the error message
#
# Results: none
#
# Side effects: see description
#
proc ::bgerror {errMsg} {
	log::log error $errMsg
	log::log error $::errorInfo
}


################################################################################
#
# Log file related procedures
#
################################################################################


# Revald_LogInit --
#
# Initialize logging
#
# Arguments: none
#
# Results: none
#
# Side effects:
# - reads the configuration file
# - starts logging
# - sets initial values for variables
#
proc Revald_LogInit {} {
	global Revald
	
	package require log
	Revald_LogSetFile
	
	# register log commands
	::log::lvCmdForall Revald_LogWrite
	::log::lvChannelForall $Revald(log_fd)
	::log::lvSuppressLE warning 0
}


# Revald_LogPreenDir --
#
# Performs cleanup in the log directory, deleting log files older than
# the configuration setting LOG_KEEP_DAYS.
#
# Arguments: none
#
# Side effects: see description
#
proc Revald_LogPreenDir {} {
	global Dir Revald
	
	if {$Revald(LOG_KEEP_DAYS) <= 0} {
		return
	}
	
	# delete old log files (may be compressed)
	foreach fileName [glob -nocomplain [file join $Dir(log) *.log*]] {
		if {[regexp {(\d{2})(\d{2})(\d{2})} $fileName date day month year]} {
			if {[catch {
					# swap the date components to satisfy [scan]
					set diffDays [expr {([clock seconds] - [clock scan $year$month$day])/86400}]
				}]} {
				log::log warning "Log file $fileName has bad time format in its file name"
			} else {
				if {$diffDays > $Revald(LOG_KEEP_DAYS)} {
					file delete $fileName
					log::log info "Deleted old log file $fileName"
				}
			}
		} else {
			log::log warning "Log file $fileName has bad time format in its file name"
		}
	}
}


# Revald_LogSetFile --
#
# This proc is called at program launch and reschedules itself to
# be run at midnight to rotate the log file. The date is used to name
# each new log file.
#
# Arguments: none
#
# Side effects:
# - opens log file for appending (create if not exists)
# - deletes deprecated log files
# - compresses previous log file if so desired
#
proc Revald_LogSetFile {} {
	global Dir Revald
	
	# reschedule this proc to be called after midnight
	set now [clock seconds]
	set next_ms [expr {([clock scan 23:59:59 -base $now] - $now + 1) * 1000}]
	after $next_ms Revald_LogSetFile
	
	# see if we need to handle a current log file
	if {[info exists Revald(logfile)] && [file exists $Revald(logfile)]} {
		set prevLogFile $Revald(logfile)
		set prevLogFd $Revald(log_fd)
		log::log notice "will rotate log (pid=[pid]) ..."
	}
	
	# open the new log file for appending. Create a new file if it doesn't
	# exist yet.
	set date [clock format $now -format %d%m%y]
	set new_logfile [file join $Dir(log) $date.log]
	if {![file isdir $Dir(log)]} {
		if {[catch {file mkdir $Dir(log)} errMsg] } {
			return -code error "Could not create log directory,\n\n$errMsg"
		}
	}
	if {[catch {
			set Revald(log_fd) [open $new_logfile a]
		} errMsg]} {
		return -code error "Could not open log file,\n\n$errMsg"
	}
	set Revald(logfile) $new_logfile
	
	# close previous log file and compress if so configured
	if {[info exists prevLogFile]} {
		catch {
			flush $prevLogFd
			close $prevLogFd
		}
		log::log notice "... continued after log rotation (pid=[pid])"
		if {$Revald(COMPRESS_CMD) ne ""} {
			if {[catch {
					eval exec $Revald(COMPRESS_CMD) $prevLogFile &
				} errMsg]} {
				log::log error \
						"Failed to compress previous log file:\n\t$errMsg"
			}
		}
	}
	
	# remove old log files if so configured
	Revald_LogPreenDir
}


# Revald_LogWrite --
#
# Writes a log message to the log in a fixed format and
# flushes the channel. Format:
#
#    level <tab> date <tab> time <tab> message
#
# Arguments:
# - level : error, warning, info or notice
# - msg   : the message to log
#
proc Revald_LogWrite {level msg} {
	global Revald
	
	set txt $level
	set now [clock seconds]
	append txt \t[clock format $now -format "%d-%b-%y\t%H:%M:%S"]
	append txt \t$msg
	puts $Revald(log_fd) $txt
	flush $Revald(log_fd)
}


#-------------------------------------------------------------------
# Start the service, and the event loop if necessary.
#-------------------------------------------------------------------
Revald_Start
if {![info exists tk_version] && ![info exists tcl_service] \
			&& ([lsearch [interp aliases] tkcon] < 0)} {
	vwait forever
}

# EOF
